{
  "type": "record",
  "name": "PlatformEvent",
  "namespace": "com.linkedin.pegasus2avro.mxe",
  "doc": "A DataHub Platform Event.",
  "fields": [
    {
      "name": "header",
      "type": {
        "type": "record",
        "name": "PlatformEventHeader",
        "doc": "A header included with each DataHub platform event.",
        "fields": [
          {
            "name": "timestampMillis",
            "type": "long",
            "doc": "The event timestamp field as epoch at UTC in milli seconds."
          }
        ]
      },
      "doc": "Header information stored with the event."
    },
    {
      "name": "name",
      "type": "string",
      "doc": "The name of the event, e.g. the type of event. For example, 'notificationRequestEvent', 'entityChangeEvent'"
    },
    {
      "name": "payload",
      "type": {
        "type": "record",
        "name": "GenericPayload",
        "doc": "Generic payload record structure for serializing a Platform Event.",
        "fields": [
          {
            "name": "value",
            "type": "bytes",
            "doc": "The value of the event, serialized as bytes."
          },
          {
            "name": "contentType",
            "type": "string",
            "doc": "The content type, which represents the fashion in which the event was serialized.\nThe only type currently supported is application/json."
          }
        ]
      },
      "doc": "The event payload."
    }
  ]
}